/***************************************************************************
 BioSequence.cc  -  description
 -------------------
 begin                : Thu Jun 16 2002
 copyright            : (C) 2002 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "BioSequence.h"
#include <string.h>
#include <sstream>
#include <stdexcept>
using namespace std;

//#define DEBUG_ALLOC
const char FRAGMENT[] = "FRAGMENT";
/** ---------------------------------------------------------- */
BioSequence::BioSequence() {
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const BioSequence& other) :
   string(other), m_sName(other.getName()), seqtype(other.seqtype) {
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const BioSequence& other): " << other.getName() << " : " << other.size() << "bp" << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const string &sSeq, const string &sName) :
   string(sSeq) {
   m_sName = sName;
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const string &sSeq, const string &sName): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const char *pcSeq, const char *pcName) :
   string(pcSeq) {
   m_sName = pcName;
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const char *pcSeq, const char *pcName): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::~BioSequence() {
#ifdef DEBUG_ALLOC
   cerr << "In ~BioSequence() for " << m_sName << endl;
#endif
}

/** ====================================== */
void BioSequence::formatedPrint(ostream &os, size_t iLineLen) const {
   size_t iPos;
   for (iPos = 0; iPos < string::size(); iPos += iLineLen) {
      os << string::substr(iPos, iLineLen) << "\n";
   }
}
/** ====================================== */
void BioSequence::unformatedPrint(ostream &os) const {
   os << c_str() << endl;
}
/** ====================================== */
void BioSequence::rmTrash() {
   string::iterator ptr = begin();
   while (ptr != end()) {
      if (!isgraph(*ptr)) {
         ptr = erase(ptr);
      } else {
         ptr++;
      }
   }
   ptr--;
   if (*ptr == '1') {
      erase(ptr);
   }
}
/** ====================================== */
void BioSequence::clear() {
   seqtype = SEQTYPE_UNKNOWN;
   setName("");
   assign("");
}

/** ====================================== */
int BioSequence::removeGaps() {
   int gap_pos = 0;
   int cnt = 0;
   BioSequence::iterator sptr = begin();
   while (sptr != end()) {
      if (isalpha(*sptr) || *sptr == '*') {
         sptr++;
      } else {
         sptr = erase(sptr);
      }
   }
   return cnt;
}
BioSequence& BioSequence::operator =(const string& s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::operator =(const char* s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::append(const BioSequence& s) {
   append((string) s);
   if (getName().size() == 0) {
      setName(s.getName());
   }
   return *this;
}
BioSequence& BioSequence::append(const string& s) {
   string::append(s);
   return *this;
}
BioSequence& BioSequence::append(const char* s) {
   string::append(s);
   return *this;
}
BioSequence& BioSequence::assign(const char* s) {
   string::assign(s);
   rmTrash();
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const char* s, size_type n) {
   string::assign(s, n);
   rmTrash();
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const char *pcSeq, const char *pcName) {
   string::assign(pcSeq);
   rmTrash();
   m_sName = pcName;
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const string& s) {
   string::assign(s);
   rmTrash();
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const string &sSeq, const string &sName) {
   string::assign(sSeq);
   rmTrash();
   m_sName = sName;
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const BioSequence &Seq) {
   string::assign(Seq);
   setName(Seq.getName());
   seqtype = Seq.seqtype;
   return *this;
}
/** ====================================== */
const string& BioSequence::getName() const {
   return m_sName;
}
/** ====================================== */
void BioSequence::setName(const string &sName) {
   m_sName = sName;
}
/** ====================================== */
void BioSequence::setName(const char *pcName) {
   m_sName = pcName;
}
/** ====================================== */
string& BioSequence::complement(string& str) {
   string::iterator rptr;
   string::iterator dptr;
   for (rptr = str.end(), rptr--, dptr = str.begin(); rptr >= dptr; rptr--, dptr++) {
      char c = *dptr;
      *dptr = complement(*rptr);
      *rptr = complement(c);
   }
   return str;
}
/** ====================================== */
BioSequence& BioSequence::complement() {
   rmTrash();
   complement(*this);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::operator =(const BioSequence &Seq) {
   assign(Seq);
   return *this;
}
/** ====================================== */
long BioSequence::findBase(long iPos) const {
   return string::find_first_not_of(GAPSYMB, iPos);
}
/** ====================================== */
BioSequence& BioSequence::append(char c) {
   string::append(1, c);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::sqeezeOutGaps() {
   string::iterator sptr, delptr;
   for (sptr = string::begin(); sptr < string::end(); sptr++) {
      if (!isalpha(*sptr)) {
         delptr = sptr;
         sptr--;
         string::erase(delptr);
      }
   }
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::format() {
   sqeezeOutGaps();
   toupper();
   return *this;
}
/** ====================================== */
void BioSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else
   printIGHead(os);
#endif

}
/** ====================================== */
void BioSequence::printIGHead(ostream& os) const {
   os << ";\n" << m_sName << "\n";
}
/** ====================================== */
void BioSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName << "\n";
}
/** ====================================== */
void BioSequence::print(ostream& os, bool format) const {
   printHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
void BioSequence::printIG(ostream& os, bool format) const {
   printIGHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
void BioSequence::printFASTA(ostream& os, bool format) const {
   printFASTAHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
char BioSequence::complement(char c) {
   //                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz"
   static const char index[] = "TVGHEFCDIJMLKNOPQYSAABWXRZ[\\]^_`tvghefcdijmlknopqysaabwxrz";
   if(c < 'A' || c > 'z') {
      return c;
   }
   return index[c-'A'];
}
/** ====================================== */
vector<char> BioSequence::string2vector(const string& s) {
   return vector<char>(s.begin(), s.end());
}
/** ====================================== */
string BioSequence::vector2string(const vector<char>& v) {
   return string(v.begin(), v.end());
}
/** ====================================== */
BioSubSequence& BioSubSequence::complement() {
   BioSequence::complement();
   size_t iTmp;
   iTmp = getFrom();
   setFrom(getTo());
   setTo(iTmp);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::toupper() {
   for (string::iterator dptr = string::begin(); dptr != string::end(); dptr++) {
      *dptr = ::toupper(*dptr);
   }
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::tolower() {
   for (string::iterator dptr = string::begin(); dptr != string::end(); dptr++) {
      *dptr = ::tolower(*dptr);
   }
   return *this;
}
/** ====================================== */
int BioSequence::read(istream& is) throw() {
   clear();
   string sLine;
   if (readHead(is)) {
      while (is.peek() != ';' && is.peek() != '>' && !(getline(is, sLine).fail())) {
         append(sLine);
      }
      rmTrash();
      guessType();
      return size();
   } else {
      return -1;
   }
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
string::const_iterator BioSequence::next_base(string::const_iterator ptr, int* gaplen) const {
   int gapnum = 0;
   while (++ptr != this->end()) {
      if (*ptr != GAPSYMB) {
         break;
      }
      gapnum++;
   }
   if (gaplen != NULL) {
      *gaplen = gapnum;
   }
   return ptr;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
string::const_iterator BioSequence::prev_base(string::const_iterator ptr, int* gaplen) const {
   int gapnum = 0;
   while (--ptr != this->begin()) {
      if (*ptr != GAPSYMB) {
         break;
      }
      gapnum++;
   }
   if (gaplen != NULL) {
      *gaplen = gapnum;
   }
   return ptr;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
char BioSequence::next_base(int p, int* gaplen) const {
   if (gaplen != NULL) {
      *gaplen = 0;
   }
   while (++p < this->size()) {
      if ((*this)[p] != GAPSYMB) {
         return (*this)[p];
      }
      if (gaplen != NULL) {
         (*gaplen)++;
      }
   }
   return 0;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
char BioSequence::prev_base(int p, int* gaplen) const {
   if (gaplen != NULL) {
      *gaplen = 0;
   }
   while (--p >= 0) {
      if ((*this)[p] != GAPSYMB) {
         return (*this)[p];
      }
      if (gaplen != NULL) {
         (*gaplen)++;
      }
   }
   return 0;
}
//==========================================================
ostream& operator <<(ostream &os, const BioSequence& seq) {
   seq.print(os);
   return os;
}
//==========================================================
istream& operator >>(istream &is, BioSequence& seq) {
   seq.read(is);
   return is;
}
/** ====================================== */
BioSubSequence::BioSubSequence() {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const BioSequence& other) : BioSequence(other) {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const BioSubSequence& other) :
   BioSequence((BioSequence) other), fragments(other.getFragments()), m_sComment(other.m_sComment) {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const string &sSeq, const string &sName, size_t iFrom, size_t iTo) :
   BioSequence(sSeq, sName), fragments(1, BioFragment(sName, iFrom, iTo)) {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const char *pcSeq, const char *pcName, size_t iFrom, size_t iTo) :
   BioSequence(pcSeq, pcName), fragments(1, BioFragment(pcName, iFrom, iTo)) {
}
/** ====================================== */
size_t BioSubSequence::getFrom() const {
   return fragments.size() ? fragments.front().first : 0;
}
/** ====================================== */
size_t BioSubSequence::getTo() const {
   return fragments.size() ? fragments.back().second : 0;
}
/** ====================================== */
const string& BioSubSequence::getComment() const {
   return m_sComment;
}
/** ====================================== */
void BioSubSequence::setFrom(size_t iFrom) {
   if (fragments.size()) {
      fragments.front().first = iFrom;
   } else {
      fragments.push_back(BioFragment(getName(), iFrom, iFrom));
   }
}
/** ====================================== */
void BioSubSequence::setTo(size_t iTo) {
   if (fragments.size()) {
      fragments.back().second = iTo;
   } else {
      fragments.push_back(BioFragment(getName(), iTo, iTo));
   }
}
/** ====================================== */
void BioSubSequence::addComment(const string& sComment) {
   m_sComment.append(sComment);
}
/** ====================================== */
void BioSubSequence::setComment(const string& sComment) {
   m_sComment = sComment;
}
/** ====================================== */
void BioSubSequence::setComment(const char *pcComment) {
   m_sComment = pcComment;
}
/** ====================================== */
void BioSubSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else
   printIGHead(os);
#endif
}
/** ====================================== */
void BioSubSequence::printIGHead(ostream& os) const {
   if (getComment().size() > 0) {
      os << "; " << getComment() << "\n";
   }
   if (getFrom() * getTo() != 0) {
      os << "; FRAGMENT " << getFrom() << " -> " << getTo() << "\n" << m_sName << "\n";
   } else {
      os << ";\n" << m_sName << "\n";
   }
}
/** ====================================== */
void BioSubSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName;
   if (getFrom() * getTo() != 0) {
      os << " FRAGMENT " << getFrom() << " -> " << getTo();
   }
   if (getComment().size() > 0) {
      os << " " << getComment();
   }
   os << "\n";
}
/** ====================================== */
BioSubSequence& BioSubSequence::assign(const BioSequence& sSeq) {
   clear();
   BioSequence::assign(sSeq);
   return *this;
}
/** ====================================== */
BioSubSequence& BioSubSequence::assign(const BioSubSequence& sSeq) {
   setFrom(sSeq.getFrom());
   setTo(sSeq.getTo());
   setComment(sSeq.getComment());
   BioSequence::assign(sSeq);
   return *this;
}
/** No descriptions */
SEQTYPE BioSequence::guessType() const {
   unsigned long piStat[BASERANGE];
   int iSize = 0;
   string::const_iterator sptr;
   for (int i = 0; i < BASERANGE; i++) {
      piStat[i] = 0;
   }
   for (sptr = begin(); sptr != end(); sptr++) {
      if (*sptr < BASERANGE) {
         if (isalpha(*sptr)) {
            piStat[::toupper(*sptr)]++;
            iSize++;
         }
      }
   }
   if (piStat['A'] + piStat['T'] + piStat['C'] + piStat['G'] > 0) {
      if ((double) (piStat['A'] + piStat['T'] + piStat['C'] + piStat['G']) / (iSize - piStat['N'] - piStat['X']) > 0.75) {
         return NA;
      }
   } else {
      if (((double) (piStat['N'])) / (iSize) > 0.75) {
         return NA;
      }
   }
   return AA;
}
/** No descriptions */
int BioSequence::readHead(istream& is) {
   string sLine;
   string sName;
   istringstream iss;
   char c;
   while (is.peek() == '\n') {
      getline(is, sLine);
   }
   if ((is.peek() == ';' || is.peek() == '>') && !(getline(is, sLine).eof())) {
      switch (sLine[0]) {
      case '>':
         iss.clear();
         iss.str(sLine);
         iss >> c >> sName;
         break;
      case ';':
         while (!(getline(is, sLine).eof()) && sLine[0] == ';')
            ;
         iss.clear();
         iss.str(sLine);
         iss >> sName;
         break;
      default:
         throw invalid_argument("BioSequence::readHead(istream*) : NO '>' or ';' FOUND.\n");
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
int BioSubSequence::readHead(istream& is) {
   string sLine;
   clear();
   while (is.peek() == '\n') {
      getline(is, sLine);
   }
   if ((is.peek() == ';' || is.peek() == '>') && !getline(is, sLine).eof()) {
      size_t iF = 0;
      size_t iT = 0;
      size_t pos;
      string sName;
      istringstream iss;
      switch (sLine[0]) {
      case '>':
         iss.str(sLine.c_str() + 1);
         iss >> sName;
         if ((pos = sLine.find(FRAGMENT)) != string::npos) {
            if (sscanf(sLine.c_str() + pos + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
               fragments.assign(1, BioFragment(sName, iF, iT));
            }
         }
         break;
      case ';':
         do {
            if ((pos = sLine.find(FRAGMENT)) != string::npos) {
               if (sscanf(sLine.c_str() + pos + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
                  fragments.assign(1, BioFragment(sName, iF, iT));
               }
            }
            sLine[0] = '\0';
         } while (!getline(is, sLine).eof() && sLine[0] == ';');
         iss.str(sLine);
         iss >> sName;
         break;
      default:
         throw invalid_argument("BioSubSequence::readHead(istream) : NO '>' or ';' FOUND.\n");
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
BioSequence& BioSubSequence::trunc_left(size_t n) {
   size_t len = ::effective_length(begin(), begin() + n);
   while(len > 0 && getFragments().size() > 0) {
      if(getFragments().front().length() <= len) {
         len -= getFragments().front().length();
         getFragments().pop_front();
      }
      else {
         if(getFragments().front().first < getFragments().front().second) {
            getFragments().front().first += len;
         }
         else {
            getFragments().front().first -= len;
         }
         len = 0;
      }
   }
   erase(0, n);
   return *this;
}
/** No descriptions */
BioSequence& BioSubSequence::trunc_right(size_t n) {
   size_t len = ::effective_length(end() - n, end());
   while(len > 0 && getFragments().size() > 0) {
      if(getFragments().back().length() <= len) {
         len -= getFragments().back().length();
         getFragments().pop_back();
      }
      else {
         if(getFragments().back().first < getFragments().back().second) {
            getFragments().back().second -= len;
         }
         else {
            getFragments().back().second += len;
         }
         len = 0;
      }
   }
   erase(size() - n);
   return *this;
}
/** No descriptions */
size_t BioSequence::effective_length() {
   return ::effective_length(begin(), end());
}
/** No descriptions */
int BioSequence::baseCount2position(int n) {
   return baseCount2position(*this, n);
}
/** No descriptions */
int BioSequence::baseCount2position(const string& s, int n) {
   int p = 0;
   for (string::const_iterator sptr = s.begin(); sptr != s.end() && n > 0; sptr++) {
      if (isalpha(*sptr)) {
         n--;
      }
      p++;
   }
   return p;
}
size_t BioSequence::baseCount(size_t pos, int n) const {
   return baseCount(*this, pos, n);
}
size_t BioSequence::baseCount(const string& s, size_t pos, int n) {
   if (pos < 0 || pos >= s.size()) {
      return 0;
   }
   int mod = n < 0 ? 1 : -1;
   size_t start = pos + n;
   if (start < 0) {
      start = 0;
   }
   if (start >= s.size()) {
      start = s.size() - 1;
   }
   size_t cnt = (isalpha(s[pos] || s[pos] == '*')) ? 1 : 0;
   for (size_t i = start; i != pos; i += mod) {
      if (isalpha(s[i] || s[i] == '*')) {
         cnt++;
      }
   }
   return cnt;
}
/** */
const BioSequence BioSequence::subseq(size_t pos, size_t n) const {
   return BioSequence(subseq(*this, pos, n), getName());
}
/** */
const BioSubSequence BioSubSequence::subseq(size_t pos, size_t n, bool literal) const {
   int sign = getFrom() <= getTo() ? +1 : -1;
   if (!literal) {
      return BioSubSequence(BioSequence::subseq(pos, n), getName(), getFrom() + sign * pos, getFrom() + sign * (pos + n));
   }
   return BioSubSequence(string::substr(pos, n), getName(), getFrom() + sign * baseCount(0, pos), getFrom() + sign * baseCount(0, pos + n));
}
/** */
const string BioSequence::subseq(const string& s, size_t pos, size_t n) {
   int beg = 0;
   int len = 0;
   while (pos + n > 0) {
      if (pos > 0) {
         beg++;
         if (isalpha(s[beg]) || s[beg] == '*') {
            pos--;
         }
      } else {
         len++;
         if (isalpha(s[beg + len]) || s[beg + len] == '*') {
            n--;
         }
      }
   }
   return s.substr(beg, len);
}
/** Read property of SEQTYPE seqtype. */
SEQTYPE BioSequence::getseqtype() {
   if (seqtype == SEQTYPE_UNKNOWN) {
      seqtype = guessType();
   }
   return seqtype;
}
string& BioSequence::shrink(string& s, char c) {
   if (s.size() < 2) {
      return s;
   }
   char lastChar = s[s.size() - 1];
   for (int i = s.size() - 2; i >= 0; i--) {
      if (s[i] == lastChar && s[i] == c) {
         s.erase(i, 1);
      } else {
         lastChar = s[i];
      }
   }
   return s;
}
void BioSequence::replaceChar(char target, char replacement) {
   for(BioSequence::iterator sptr = begin(); sptr != end(); sptr++) {
      if(*sptr == target) {
         *sptr = replacement;
      }
   }
}
/*
 * Modifies coordinates relative to the source sequence.
 * [start, end] - coordinates relative to this subsequence.
 * **/
void BioSubSequence::adjust_position(int start, int end) {
   if (getFragments().size() == 0) {
      getFragments().assign(1, BioFragment(getName(), start, end));
      return;
   }
   for(list<BioFragment>::iterator lptr = getFragments().begin(); lptr != getFragments().end(); lptr++) {
      if((*lptr).first <= (*lptr).second) {
         (*lptr).second = (*lptr).first + (end - 1);
         (*lptr).first  = (*lptr).first + (start - 1);
      }
      else {
         (*lptr).second = (*lptr).first - (end - 1);
         (*lptr).first  = (*lptr).first - (start - 1);
      }
   }
}
BioSubSequence& BioSubSequence::append(BioSubSequence& other) {
   BioSequence::append((BioSequence) other);
   if (other.getFragments().size() > 0) {
      getFragments().insert(getFragments().end(), other.getFragments().begin(), other.getFragments().end());
   }
   return *this;
}
/** ====================================== */
void BioSubSequence::clear() {
   BioSequence::clear();
   setFrom(0);
   setTo(0);
   setComment("");
   getFragments().clear();
}

RatedSequence::RatedSequence() :
   score(0) {
}
RatedSequence::RatedSequence(const BioSequence& other) :
   BioSequence(other), score(0) {
}
RatedSequence::RatedSequence(const RatedSequence& other) :
   BioSequence(other), score(other.score) {
}
RatedSequence& RatedSequence::assign(const RatedSequence& other) {
   BioSequence::assign((BioSequence) other);
   score = other.score;
   return *this;
}
void RatedSequence::setScore(double score) {
   this->score = score;
}
double RatedSequence::getScore() const {
   return score;
}
void RatedSequence::clear() {
   BioSequence::clear();
   score = 0;
}
/** ====================================== */
void RatedSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else

   printIGHead(os);
#endif

}
/** ====================================== */
void RatedSequence::printIGHead(ostream& os) const {
   if (score) {
      os << "; SCORE= " << getScore() << endl;
   }
   BioSequence::printIGHead(os);
}
/** ====================================== */
void RatedSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName;
   if (score) {
      os << " SCORE= " << getScore();
   }
   os << endl;
}
